package com.persagy.dmp.rwd.migrate.service.impl;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.http.HttpUtil;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.persagy.dmp.common.context.AppContext;
import com.persagy.dmp.common.exception.BusinessException;
import com.persagy.dmp.mybatis.helper.DynamicDataSourceHelper;
import com.persagy.dmp.rwd.migrate.dao.DigitalMigrateMapper;
import com.persagy.dmp.rwd.migrate.service.IDigitalMigrateService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * 数字孪生 数据迁移
 * @author Charlie Yu
 * @date 2021-08-04
 */
@Service
@Transactional(rollbackFor = Exception.class)
@Slf4j
public class DigitalMigrateServiceImpl implements IDigitalMigrateService {

    @Resource
    private DigitalMigrateMapper dao;

    @Override
    public void startMigrate(List<String> groupCodes, List<String> projectIds, String orgUrl, String oldSchema) {
        Set<String> groupCodeSet = queryGroupCodes(orgUrl, groupCodes);
        if(CollUtil.isEmpty(groupCodeSet)) {
            log.error("没有查询到集团信息！");
            return;
        }
        // 按集团循环处理
        for(String groupCode:groupCodeSet) {
            log.info(StrUtil.format("开始处理{}集团", groupCode));
            // 获取待处理的项目
            Set<String> projectIdSet = queryProjectIds(orgUrl, groupCode, projectIds);
            if(CollUtil.isEmpty(projectIdSet)) {
                log.info(StrUtil.format("结束处理{}集团：没有找到需要迁移的项目", groupCode));
                continue;
            }
            // 集团数据迁移
            doMigrate(oldSchema, groupCode);
            // 项目数据迁移
            for(String projectId:projectIdSet) {
                doMigrate(oldSchema, groupCode, projectId);
            }
            log.info(StrUtil.format("结束处理{}集团：迁移完成!!!", groupCode));
        }
    }

    /**
     * 执行集团级数据迁移
     * @param schema 数据中台库实例
     * @param groupCode 集团编码
     */
    private void doMigrate(String schema, String groupCode) {
        // 切换集团数据源
        AppContext.getContext().setGroupCode(groupCode);
        DynamicDataSourceHelper.loadDataSource();
        // 变小写
        groupCode = groupCode.toLowerCase();
        // 迁移
        log.info(StrUtil.format("正在处理{}集团：迁移类型定义数据...", groupCode));
        dao.migrateDefineType(schema, groupCode);
        log.info(StrUtil.format("正在处理{}集团：迁移信息点定义数据...", groupCode));
        dao.migrateDefineInfo(schema, groupCode);
        log.info(StrUtil.format("正在处理{}集团：迁移图定义数据...", groupCode));
        dao.migrateDefineGraph(schema, groupCode);
        log.info(StrUtil.format("正在处理{}集团：迁移边定义数据...", groupCode));
        dao.migrateDefineRel(schema, groupCode);
    }

    /**
     * 执行项目级迁移
     * @param schema 数据中台库实例
     * @param groupCode 集团编码
     * @param projectId 项目ID
     */
    private void doMigrate(String schema, String groupCode, String projectId) {
        // 已经切换到集团数据源了，不需要再切换了
        groupCode = groupCode.toLowerCase();
        projectId = projectId.toLowerCase();
        // 迁移
        log.info(StrUtil.format("正在处理{}集团{}项目：迁移对象数据...", groupCode, projectId));
        dao.migrateObject(schema, groupCode, projectId);
        log.info(StrUtil.format("正在处理{}集团{}项目：迁移关系数据...", groupCode, projectId));
        dao.migrateRelation(schema, groupCode, projectId);
    }

    /**
     * 查询项目主键
     * @param orgUrl 组织服务url
     * @param groupCode 集团编码，非空
     * @param ids 指定的项目ID，为空表示所有
     * @return
     */
    private Set<String> queryProjectIds(String orgUrl, String groupCode, List<String> ids) {
        // 调用数据中台接口查询项目信息
        String queryProjectUrl = StrUtil.format(orgUrl + "/org/project/query?groupCode={}&userId=systemId", groupCode);
        String response = HttpUtil.post(queryProjectUrl, "{}");
        JSONArray projects = fetchResult(response);
        if(CollUtil.isEmpty(projects)) {
            return null;
        }
        // 收集项目ID
        Set<String> projectIds = new HashSet<>();
        for(int i = 0;i < projects.size();i++) {
            JSONObject project = projects.getJSONObject(i);
            String id = project.getString("id");
            boolean valid = project.getBooleanValue("status");
            // 已删除或没有编码的不处理
            if(StrUtil.isBlank(id) || !valid) {
                continue;
            }
            projectIds.add(id);
        }
        // 处理返回结果
        if(CollUtil.isEmpty(projectIds)) {
            return null;
        } else if(CollUtil.isNotEmpty(ids)) {
            // 有指定的集团，则取交集
            return CollUtil.intersectionDistinct(projectIds, ids);
        } else {
            // 没有指定集团，则返回所有有效集团
            return projectIds;
        }
    }

    /**
     * 查询集团编码
     * @param orgUrl 组织服务url
     * @param codes 指定集团编码 - 为空表示所有集团
     * @return
     */
    private Set<String> queryGroupCodes(String orgUrl, List<String> codes) {
        // 调用数据中台接口查询集团信息
        String queryGroupUrl = orgUrl + "/org/group/query";
        String groupResponse = HttpUtil.post(queryGroupUrl, "{}");
        JSONArray groups = fetchResult(groupResponse);
        if(CollUtil.isEmpty(groups)) {
            return null;
        }
        // 收集集团编码
        Set<String> groupCodes = new HashSet<>();
        for(int i = 0;i < groups.size();i++) {
            JSONObject group = groups.getJSONObject(i);
            String code = group.getString("code");
            boolean valid = group.getBooleanValue("status");
            // 已删除或没有编码的不处理
            if(StrUtil.isBlank(code) || !valid) {
                continue;
            }
            groupCodes.add(code);
        }
        // 处理返回结果
        if(CollUtil.isEmpty(groupCodes)) {
            return null;
        } else if(CollUtil.isNotEmpty(codes)) {
            // 有指定的集团，则取交集
            return CollUtil.intersectionDistinct(groupCodes, codes);
        } else {
            // 没有指定集团，则返回所有有效集团
            return groupCodes;
        }
    }

    /**
     * 提取结果
     * @param response
     * @return
     */
    private JSONArray fetchResult(String response) {
        JSONObject responseObject = JSONObject.parseObject(response);
        // 如果结果不成功，抛出失败结果
        if(!StrUtil.equals("success", responseObject.getString("result"))) {
            throw new BusinessException(responseObject.getString("message"));
        }
        JSONArray datas = responseObject.getJSONArray("data");
        return CollUtil.isEmpty(datas) ? null : datas;
    }
}
